from pprint import pprint

from scmdata import ScmRun


def check_all_scenarios_as_in_protocol(df_to_check, protocol_scenarios, **kwargs):
    checker_df = df_to_check["scenario"].to_frame()
    merged_df = checker_df.merge(protocol_scenarios[["scenario"]])
    assert len(merged_df) == len(checker_df), set(checker_df["scenario"]) - set(
        merged_df["scenario"]
    )
    return df_to_check


def check_all_variables_and_units_as_in_protocol(
    df_to_check, protocol_variables, **kwargs
):
    checker_df = df_to_check.filter(variable="*Other*", keep=False).meta[
        ["variable", "unit"]
    ]
    checker_df["unit"] = checker_df["unit"].apply(
        lambda x: x.replace("dimensionless", "Dimensionless")
        if isinstance(x, str)
        else x
    )

    def strip_quantile(inv):
        if any([inv.endswith(suf) for suf in ["quantile", "mean", "stddev"]]):
            return "|".join(inv.split("|")[:-1])

        return inv

    checker_df["variable"] = checker_df["variable"].apply(strip_quantile)
    merged_df = checker_df.merge(protocol_variables[["variable", "unit"]])
    try:
        assert len(merged_df) == len(checker_df)
    except AssertionError:  # pragma: no cover
        pprint(set(checker_df["variable"]) - set(protocol_variables["variable"]))
        pprint(set(checker_df["unit"]) - set(protocol_variables["unit"]))
        raise


def unify_units(in_df, protocol_variables, **kwargs):
    out_df = in_df.copy()

    def get_units(v):
        try:
            return protocol_variables[protocol_variables["variable"] == v]["unit"].iloc[
                0
            ]
        except Exception:
            raise AssertionError(f"Failed to find unit for {variable}")

    for variable in out_df["variable"].unique():
        if variable.startswith("Radiative Forcing|Anthropogenic|Albedo Change"):
            target_unit = get_units("Radiative Forcing|Anthropogenic|Albedo Change")

        elif variable.startswith(
            "Effective Radiative Forcing|Anthropogenic|Albedo Change"
        ):
            target_unit = get_units(
                "Effective Radiative Forcing|Anthropogenic|Albedo Change"
            )

        elif variable.startswith("Carbon Pool"):
            target_unit = get_units("Carbon Pool|Atmosphere")

        elif "Other" in variable:
            target_unit = get_units("{}".format(variable.split("|Other")[0]))

        elif any([variable.endswith(suf) for suf in ["quantile", "mean", "stddev"]]):
            target_unit = get_units("|".join(variable.split("|")[:-1]))
        else:
            target_unit = get_units(variable)

        try:
            if "CH4" in target_unit:
                out_df = out_df.convert_unit(
                    target_unit, variable=variable, context="CH4_conversions"
                )
            elif "NOx" in target_unit:
                out_df = out_df.convert_unit(
                    target_unit, variable=variable, context="NOx_conversions"
                )
            else:
                if target_unit == "Dimensionless":
                    target_unit = "dimensionless"

                out_df = out_df.convert_unit(target_unit, variable=variable)
        except Exception:  # pragma: no cover
            current_unit = out_df.filter(variable=variable)["unit"].unique()
            raise AssertionError(
                f"Failed for {variable} with target unit: {target_unit} and current_unit: {current_unit}"
            )

    out_df = out_df.timeseries().reset_index()
    out_df["unit_context"] = out_df["unit_context"].fillna("not_required")
    return ScmRun(out_df)


all_checks = [
    ("fix_units", unify_units),
    ("check_scenarios", check_all_scenarios_as_in_protocol),
    ("check_variables", check_all_variables_and_units_as_in_protocol),
]



import pandas as pd
import pkg_resources
from scmdata.run import ScmRun, df_append

SUBMISSION_TEMPLATE = "data/rcmip-data-submission-template-v4-0-0.xlsx"


def read_results_csvs(results):
    """

    Parameters
    ----------
    results: str or list of str
        Files to read in. All files to be read should be formatted as CSV or XLSX files following the formatting
        defined in the template spreadsheet.

    Returns
    -------
        `scmdata.ScmDataFrame
    """
    if isinstance(results, str):
        results = [results]

    db = []
    for rf in results:
        if rf.endswith(".csv"):
            loaded = ScmRun(rf)
        else:
            loaded = ScmRun(rf, sheet_name="your_data")
        db.append(loaded)

    db = df_append(db).timeseries().reset_index()
    db = ScmRun(db)
    return db


def read_protocol_variables():
    stream = pkg_resources.resource_stream(__name__, SUBMISSION_TEMPLATE)
    protocol_variables = pd.read_excel(stream, sheet_name="variable_definitions")
    protocol_variables.columns = protocol_variables.columns.str.lower()

    return protocol_variables


def read_protocol_scenarios():
    stream = pkg_resources.resource_stream(__name__, SUBMISSION_TEMPLATE)
    protocol_scenarios = pd.read_excel(stream, sheet_name="scenario_info")
    protocol_scenarios.columns = protocol_scenarios.columns.str.lower()

    return protocol_scenarios



import versioneer
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

PACKAGE_NAME = "pyrcmip"
DESCRIPTION = "Tools for accessing RCMIP data"
KEYWORDS = ["data", "simple climate model", "climate", "scm"]

AUTHORS = [
    ("Zeb Nicholls", "zebedee.nicholls@climate-energy-college.org"),
    ("Jared Lewis", "jared.lewis@climate-energy-college.org"),
]
EMAIL = "zebedee.nicholls@climate-energy-college.org"
URL = "https://gitlab.com/rcmip/pyrcmip"
PROJECT_URLS = {
    "Bug Reports": "https://github.com/rcmip/pyrcmip/issues",
    # "Documentation": "https://rcmip.readthedocs.io/en/latest",
    "Source": "https://gitlab.com/rcmip/pyrcmip",
}
LICENSE = "3-Clause BSD License"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: BSD License",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]

REQUIREMENTS = ["click", "scmdata", "boto3", "pyjwt", "semver", "xlrd", "docutils<0.16"]
REQUIREMENTS_TESTS = [
    "codecov",
    "nbval",
    "pytest-cov",
    "pytest-mock",
    "pytest>=5.0.0",
    "moto",
]
REQUIREMENTS_DOCS = ["sphinx>=1.4,<2.1", "sphinx_rtd_theme"]
REQUIREMENTS_DEPLOY = ["twine>=1.11.0", "setuptools>=38.6.0", "wheel>=0.31.0"]

REQUIREMENTS_DEV = [
    *["flake8", "isort", "black", "pydocstyle"],
    *REQUIREMENTS_TESTS,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_DEPLOY,
]

REQUIREMENTS_EXTRAS = {
    "docs": REQUIREMENTS_DOCS,
    "tests": REQUIREMENTS_TESTS,
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": REQUIREMENTS_DEV,
}

SOURCE_DIR = "src"

PACKAGES = find_packages(SOURCE_DIR)  # no exclude as only searching in `src`
PACKAGE_DIR = {"": SOURCE_DIR}
PACKAGE_DATA = {"": ["data/*.xlsx"]}

ENTRY_POINTS = {"console_scripts": ["rcmip = pyrcmip.cli:run_cli"]}

README = "README.rst"

with open(README, "r") as readme_file:
    README_TEXT = readme_file.read()


class RCMIP(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": RCMIP})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    author=", ".join([author[0] for author in AUTHORS]),
    author_email=", ".join([author[1] for author in AUTHORS]),
    url=URL,
    project_urls=PROJECT_URLS,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
    entry_points=ENTRY_POINTS,
)



from logging import getLogger

from scmdata.run import ScmRun, df_append

from pyrcmip.readers import (
    read_protocol_scenarios,
    read_protocol_variables,
    read_results_csvs,
)

from .checks import all_checks

logger = getLogger(__name__)


def run_validation(files):
    try:
        results_df = read_results_csvs(files)
    except ValueError:
        logger.exception("Could not load data")
        raise ValueError("Data failed validation. See log for more details")

    protocol_variables = read_protocol_variables()
    protocol_scenarios = read_protocol_scenarios()

    clean_db = []
    any_failures = False

    for climatemodel, cdf in results_df.timeseries().groupby("climatemodel"):
        logger.info("checking results for climate model {}".format(climatemodel))

        any_failures_climatemodel = False

        cdf = ScmRun(cdf)

        try:
            for func_name, check in all_checks:
                logger.info(func_name)
                res = check(
                    cdf,
                    protocol_variables=protocol_variables,
                    protocol_scenarios=protocol_scenarios,
                )
                if res is not None:
                    cdf = res
        except AssertionError as e:
            logger.error(e)
            any_failures_climatemodel = True
        #     # currently not possible as groups weren't told to obey variable hierarchy,
        #     # add this in phase 2
        #     for v_top in cdf_converted_units.filter(level=0)["variable"].unique():
        #         print(v_top)
        #         cdf_pyam = cdf_converted_units.filter(variable="{}*".format(v_top)).timeseries()
        #         cdf_pyam.columns = cdf_pyam.columns.map(lambda x: x.year)

        #         cdf_consistency_checker = pyam.IamDataFrame(cdf_pyam)
        #         if cdf_consistency_checker.check_internal_consistency() is not None:
        #             print("Failed for {}".format(v_top))
        #             any_failures_climatemodel = True
        #             failing_set = cdf_consistency_checker.copy()

        if not any_failures_climatemodel:
            clean_db.append(cdf)
            logger.info("All clear for {}".format(climatemodel))
        else:
            logger.error("Failed {}".format(climatemodel))
            any_failures = True
    if any_failures:
        logger.error("database isn't ready yet")
        raise ValueError("Data failed validation. See log for more details")
    else:
        return df_append(clean_db)



# -*- coding: utf-8 -*-

"""Console script for rcmip."""
import logging
import sys
from glob import glob
from os.path import basename, join

import boto3
import click
import jwt
import semver
from botocore.exceptions import ClientError

from pyrcmip.validate import run_validation

logger = logging.getLogger("rcmip.cli")


class ColorFormatter(logging.Formatter):
    colors = {
        "error": dict(fg="red"),
        "exception": dict(fg="red"),
        "critical": dict(fg="red"),
        "debug": dict(fg="blue"),
        "warning": dict(fg="yellow"),
    }

    def format(self, record):
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.getMessage()
            if level in self.colors:
                prefix = click.style("{}: ".format(level), **self.colors[level])
                msg = "\n".join(prefix + x for x in msg.splitlines())
            return msg
        return logging.Formatter.format(self, record)


class ClickHandler(logging.Handler):
    _use_stderr = True

    def emit(self, record):
        try:
            msg = self.format(record)
            click.echo(msg, err=self._use_stderr)
        except Exception:  # pragma: no cover
            self.handleError(record)


_default_handler = ClickHandler()
_default_handler.formatter = ColorFormatter()


@click.group(name="rcmip")
@click.option("--log-level", default="INFO")
def cli(log_level):
    root = logging.getLogger()
    root.handlers.append(_default_handler)
    root.setLevel(log_level)


def clean_input_fnames(ctx, param, inp):
    filenames = []

    for f in inp:
        filenames.extend(glob(f))

    for f in filenames:
        logger.info("found {}".format(f))

    if len(filenames) == 0:
        raise click.BadParameter("No valid input files found")

    return sorted(set(filenames))


@cli.command()
@click.argument("input", nargs=-1, type=click.Path(), callback=clean_input_fnames)
def validate(input):
    try:
        run_validation(input)
    except Exception as e:
        raise click.ClickException(str(e))


def validate_version(ctx, param, value):
    try:
        s = semver.VersionInfo.parse(value)

        if s.prerelease is None and s.build is None:
            return value
        else:
            raise click.BadParameter(
                "Version must only contain major, minor and patch values"
            )
    except ValueError:
        raise click.BadParameter("Cannot parse version string")


@cli.command()
@click.option("--validate/--no-validate", default=True)
@click.option(
    "--token",
    required=True,
    help="Authentication token. Contact zebedee.nicholls@climate-energy-college.org for a token",
)
@click.option("--bucket", default="rcmip-uploads-au")
@click.option("--model", required=True)
@click.option(
    "--version",
    required=True,
    callback=validate_version,
    help="Version of the data being uploaded. Must be a valid semver version string (https://semver.org/). "
    "For example 2.0.0",
)
@click.argument("input", nargs=-1, type=click.Path(), callback=clean_input_fnames)
def upload(validate, token, bucket, model, version, input):
    if validate:
        try:
            run_validation(input)
        except Exception:
            raise click.ClickException("Validation failed. Fix issues and rerun")

    # Upload data to S3
    t = jwt.decode(token, verify=False)
    session = boto3.session.Session(
        aws_access_key_id=t["access_key_id"],
        aws_secret_access_key=t["secret_access_key"],
    )
    client = session.client("s3")

    root_key = "{}/{}/{}".format(t["org"], model, version)

    # Check if this version is already uploaded (using the {key}-complete dummy file)
    try:
        client.head_object(Bucket=bucket, Key=root_key + "-complete")

        raise click.ClickException(
            "Data for this version has already been uploaded. Increment the version and try again"
        )
    except ClientError:
        logger.debug("Object with key {} does not exist".format(root_key))

    try:
        for f in input:
            key = join(root_key, basename(f))
            logger.info("Uploading {}".format(f))
            client.upload_file(Bucket=bucket, Key=key, Filename=f)
    except ClientError:  # pragma: no cover
        logger.exception("Failed to upload file")
        raise click.ClickException("Failed to upload file")

    # Finally mark the upload as complete by uploading a dummy file
    # Writing this dummy file will be used to start the processing of the upload
    client.put_object(Bucket=bucket, Key=root_key + "-complete")

    logger.info("All files uploaded successfully")


def run_cli():
    sys.exit(cli(auto_envvar_prefix="RCMIP"))  # pragma: no cover


if __name__ == "__main__":
    run_cli()  # pragma: no cover
